# -*- coding: utf-8 -*-
import json
import string
import hashlib
import logging
import hmac
from hashlib import sha256

from datetime import date
from dateutil.relativedelta import relativedelta

from odoo import Command, models, fields, api
from odoo.osv import expression
from odoo.tools import float_utils
from odoo.exceptions import UserError
from odoo.addons.saas_worker.const import IAP_FEE, APPS_KEY, APPS_URL, APPS_FEE

_logger = logging.getLogger(__name__)


# --------------------------------------
#  create/update IAP/APPS product
# --------------------------------------
class ProductTemplate(models.Model):
    _inherit = "product.template"

    apps_reference = fields.Char(string='Apps reference', compute='_compute_apps_reference')

    # product_variant_id depends on product_variant_ids, and takes the first item of the collection
    # this compute method will also recompute apps_reference for product_variant_ids that aren't
    # the first of the list, but we don't really mind, because apps_reference is usually set
    # on products that have only one variant.
    # using a related= instead of a compute= causes a warning since 16.0
    @api.depends('product_variant_ids.apps_reference')
    def _compute_apps_reference(self):
        for product_template in self:
            product_template.apps_reference = product_template.product_variant_id.apps_reference

    def _set_all_companies_service_to_purchase(self, service_to_purchase):
        """
        Set the company-dependent property service_to_purchase to the same value on all companies
        """
        self.ensure_one()

        IrProperty = self.env['ir.property']
        field_id = self.env['ir.model.fields']._get('product.template', 'service_to_purchase')

        res_id = 'product.template,%d' % self.id
        props = IrProperty.search([
            ('fields_id', '=', field_id.id),
            ('res_id', '=', res_id),
        ])
        default_prop = props.filtered(lambda x: x.company_id is False)[:1]

        # remove company-specific values
        (props - default_prop).unlink()

        # update or create the all-company value
        if default_prop:
            default_prop.write({
                'value': service_to_purchase,
            })
        else:
            IrProperty.create({
                'fields_id': field_id.id,
                'company_id': False,
                'res_id': res_id,
                'name': 'service_to_purchase',
                'value': service_to_purchase,
                'type': self._fields['service_to_purchase'].type,
            })


class ProductProduct(models.Model):
    _inherit = "product.product"

    apps_reference = fields.Char(string='Apps reference')
    digital_content = fields.Boolean('Digital Content', default=False, help="If checked, it will allow clients to download the product attachments when they have bought it.")

    def _is_add_to_cart_allowed(self):
        self.ensure_one()

        allowed_categories_xmlids = [
            'openerp_apps.apps_product_category', # 3rd party apps
            'openerp_apps.iap_product_category',  # IAP products
        ]

        return super()._is_add_to_cart_allowed() or self.sudo().categ_id in [self.env.ref(xmlid) for xmlid in allowed_categories_xmlids]

    def _temp_dl_link(self, days=7):
        """ this function is made public in order to be called in mail/qweb renders.
        """
        self.ensure_one()
        validity_date = date.today() + relativedelta(days=days)
        message = 'apps-download-token-%s-%s-%s' % (validity_date, self.default_code, self.id)
        hm = hmac.new(APPS_KEY.encode(), message.encode(), sha256).hexdigest()
        url = '{host}/loempia/download/tmp/{module}/%s/%s/%s.zip' % (validity_date, self.id, hm)
        return url.format(host=APPS_URL, module=self.default_code)

    @api.model
    def _iap_product_update(self, values):
        cat = self.env.ref('openerp_apps.iap_product_category')
        public_cat = self.env.ref('openerp_apps.iap_public_product_category')
        product_fee = IAP_FEE
        service_to_purchase = True

        return self._product_update(cat, public_cat, product_fee, service_to_purchase, values)

    @api.model
    def _apps_product_update(self, values):
        cat = self.env.ref('openerp_apps.apps_product_category')
        public_cat = self.env.ref('openerp_apps.apps_public_product_category')
        product_fee = APPS_FEE
        service_to_purchase = True

        return self._product_update(cat, public_cat, product_fee, service_to_purchase, values)

    @api.model
    def _product_update(self, cat, public_cat, product_fee, service_to_purchase, values):
        """
            :param cat: product.category object
            :param public_cat: product.public.category object
            :param product_fee: percentage of commission on the product
            TODO RAR - legacy: following parameter will be removed once we change to the automatic PO generation
            :param service_to_purchase: boolean
            :params values: dictionary with keys
                    - name*
                    - list_price*
                    - default_code*
                    - oauth_uid*
                    - apps_reference*
                    - description_sale
                    - digital_content
                    - image
                    - active
                    - website_published
                    * required for both iap and apps
        """

        if cat and public_cat:
            values.update({
                'categ_id': cat.id,
                'public_categ_ids': [(6, 0, [public_cat.id])],
            })

        # Automatically set VAT for product
        ICP = self.env['ir.config_parameter'].sudo()
        vat_ids = json.loads(ICP.get_param('openerp_apps.apps_product_tax', '[]'))
        supp_vat_ids = json.loads(ICP.get_param('openerp_apps.apps_seller_tax', '[]'))
        if not (vat_ids and supp_vat_ids):
            raise UserError('IAP/Apps Taxes could not be found.')

        values.update(
            {
                'service_to_purchase': False,
                'purchase_method': 'purchase',
                'purchase_ok': True,
                'sale_ok': False,
                'type': 'service',
                'taxes_id': [Command.set(vat_ids)],
                'supplier_taxes_id': [Command.set(supp_vat_ids)],
                'available_in_pos': False,
            }
        )

        oauth_uid = values.pop('oauth_uid', '0')
        product_id = values.pop('product_id', False)
        apps_reference = values.get('apps_reference', '')

        if 'image' in values:
            values['image_1920'] = values.pop('image').encode()
        # Update or create corresponding product
        # legacy issue: We look for product that already exist in prod but that don't have apps_reference defined.
        # As long as we support that modules are bought for all versions, this product will be related to the latest version.
        # On Apps, however, all modules from all versions point towards the same product_id.
        # For the time being, we can look for the product_id. Once we drop the 'all versions' purchase,
        # we don't want to overwrite the apps_reference field of a product that was priorly linked to the latest version
        # --> hence look for matching apps_reference fields.
        product = None
        if product_id:
            #TODO: drop this condition along with 'all versions' support.
            product = self.browse(product_id)

        if not (product and product.exists()):
            domain = [
                '&',
                    ('categ_id', '=', cat.id),
                    '|',
                        ('apps_reference', '=', apps_reference),
                        # The product_id can only be trusted if there is no apps_reference on the product,
                        # otherwise it might not have been synced on apps yet.
                        '&',
                            ('apps_reference', '=', False),
                            ('id', '=', product_id)
            ] # noqa
            product = self.search(domain, limit=1, order='create_date asc')
        if not product.exists():
            product = self.create(values)
        else:
            # avoid triggering dependent computed fields, such as l10n_in_hsn_code on account.move.line
            # indirectly impacted because l10n_in_hsn_code comes from the category too
            # (cfr dd9267de0257afbf0459fcdd0070b32c87bd1c59 + odoo/odoo@8d6a59da4f62c440c22d677bad6b70b941e25ea6)
            if product.categ_id == cat:
                values.pop('categ_id', None)

            product.write(values)
            # update the `active` value only when needed.
            # On the `product` module, if the `active` field is written
            # then the cache is cleared.
            # see PR: https://github.com/odoo/odoo/pull/30274/files#diff-096fa28ccebc1853fa4b84107be9800b76547b5fbc52a9c751b7c969d098eb4bR358
            new_active_value = values.get('active', True)
            if product.product_tmpl_id.active != new_active_value:
                product.product_tmpl_id.active = new_active_value

        # if product_unspsc is installed (for l10n_co, l10n_mx or l10n_pe), products need to have a specific category set
        if hasattr(product, 'unspsc_code_id') and not product.unspsc_code_id:
            product.unspsc_code_id = self.env.ref('product_unspsc.unspsc_code_43231602')

        # updates a supplierinfo object to link to the product or creates it if inexistent or if the product list_price has changed.
        publisher_user_id = self.env['res.users'].browse(int(oauth_uid))
        if not publisher_user_id.exists():
            _logger.error('IAP/Apps: An error occured during the product creation. The creator\'s identity is missing.')
            return 0, 0

        apps_vendor_id = self.env.ref('openerp_apps.apps_vendor_partner_category')
        partner = publisher_user_id.partner_id
        partner.write({'category_id': [(4, apps_vendor_id.id)]})

        supplier_precision = self.env['decimal.precision'].precision_get('Product Price') or 2
        supplier_price = float_utils.float_round(product.list_price * (1 - product_fee), precision_digits=supplier_precision)
        supplier = product.seller_ids and product.seller_ids[-1]

        if not supplier\
            or float_utils.float_compare(supplier.price, supplier_price, precision_digits=supplier_precision)\
            or supplier.partner_id != partner\
            or supplier.product_id != product:

            tmpl_id = product.product_tmpl_id
            vals = {
                'currency_id': tmpl_id.currency_id.id,
                'product_tmpl_id': tmpl_id.id,
                'product_id': product.id,
                'partner_id': partner.id,
                'company_id': False,
                'price': supplier_price,
                'sequence': product.seller_ids and max(product.seller_ids.mapped('sequence')) + 1 or 1,
            }
            supplier = self.env['product.supplierinfo'].create(vals)
            _logger.info('openerp_apps: Creating a supplier for product %s (apps_reference: %s) with values %s', product.id, apps_reference, str(vals))
        else:
            _logger.info('openerp_apps: Using supplier %r for product %r', supplier, product)

        # Update the service_to_purchase field after the supplier creation or update
        product.product_tmpl_id._set_all_companies_service_to_purchase(service_to_purchase)

        return product.id, supplier.id


class ProductSupplierinfo(models.Model):
    _inherit = "product.supplierinfo"

    # Ignore supplierinfo from searches if its partner is disabled
    active = fields.Boolean(related='partner_id.active', readonly=True)
